/**
 * @brief Kqueue Abstraction Layer (KAL) for BSD
 *
 * @file sys_kqueue.c
 */
/* -----------------------------------------------------------------------------
 * Enduro/X Middleware Platform for Distributed Transaction Processing
 * Copyright (C) 2009-2016, ATR Baltic, Ltd. All Rights Reserved.
 * Copyright (C) 2017-2018, Mavimax, Ltd. All Rights Reserved.
 * This software is released under one of the following licenses:
 * AGPL or Mavimax's license for commercial use.
 * -----------------------------------------------------------------------------
 * AGPL license:
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License, version 3 as published
 * by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Affero General Public License, version 3
 * for more details.
 *
 * You should have received a copy of the GNU Affero General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * -----------------------------------------------------------------------------
 * A commercial use license is available from Mavimax, Ltd
 * contact@mavimax.com
 * -----------------------------------------------------------------------------
 */

/*---------------------------Includes-----------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <unistd.h>
#include <stdarg.h>
#include <ctype.h>
#include <memory.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <pthread.h>
#include <string.h>
#include <sys/event.h>

#include <ndrstandard.h>
#include <ndebug.h>
#include <nstdutil.h>
#include <limits.h>

#include <sys_unix.h>


/*---------------------------Externs------------------------------------*/
/*---------------------------Macros-------------------------------------*/
/*---------------------------Enums--------------------------------------*/
/*---------------------------Typedefs-----------------------------------*/
/*---------------------------Globals------------------------------------*/
/*---------------------------Statics------------------------------------*/
/*---------------------------Prototypes---------------------------------*/

/**
 * Nothing to init for epoll()
 */
expublic int ndrx_epoll_sys_init(void)
{
    return EXSUCCEED;
}

/**
 * Nothing to un-init for epoll()
 */
expublic inline void ndrx_epoll_sys_uninit(void)
{
    return;
}

/**
 * Return the compiled poll mode
 * @return 
 */
expublic inline char * ndrx_epoll_mode(void)
{
    static char *mode = "kqueue";
    
    return mode;
}
/**
 * Wrapper for epoll_ctl, for standard file descriptors
 * @param epfd
 * @param op
 * @param fd
 * @param event
 * @return 
 */
expublic inline int ndrx_epoll_ctl(int epfd, int op, int fd, struct ndrx_epoll_event *event)
{   
    int ret = EXSUCCEED;
    char *fn = "ndrx_epoll_ctl";
    struct kevent ev;
    
    ev.udata = NULL;

    if (EX_EPOLL_CTL_ADD == op)
    {
        NDRX_LOG(log_info, "%s: Add operation on ndrx_epoll set %d, fd %d", fn, epfd, fd);
        
        EV_SET(&ev, fd, event->events, EV_ADD, 0, 0, NULL);
        
        return kevent(epfd, &ev, 1, NULL, 0, NULL);
    }
    else if (EX_EPOLL_CTL_DEL == op)
    {
        NDRX_LOG(log_info, "%s: Delete operation on ndrx_epoll set %d, fd %d", fn, epfd, fd);
           
        EV_SET(&ev,fd, EVFILT_READ, EV_DELETE, 0, 0, NULL);
        
        return kevent(epfd, &ev, 1, NULL, 0, NULL);
    }
    else
    {
        NDRX_LOG(log_error, "Invalid operation %d", op);
	errno = EINVAL;
        EXFAIL_OUT(ret);
    }
    
out:

    NDRX_LOG(log_info, "%s return %d", fn, ret);

    return ret;
}

/**
 * epoll_ctl for Posix queue descriptors
 * @param epfd
 * @param op
 * @param fd
 * @param event
 * @return 
 */
expublic inline int ndrx_epoll_ctl_mq(int epfd, int op, mqd_t fd, struct ndrx_epoll_event *event)
{
    int ret = EXSUCCEED;
    char *fn = "ndrx_epoll_ctl";
    struct kevent ev;
    void *tmp = fd;
    int real_fd = *((int *)tmp);

    /* for bsd the file descriptor must be dereferenced from pointer */

    if (EX_EPOLL_CTL_ADD == op)
    {
        NDRX_LOG(log_info, "%s: Add operation on ndrx_epoll set %d, fd %d", fn, epfd, fd);

        EV_SET(&ev, real_fd, EVFILT_READ, event->events, 0, 0, NULL);
        ev.udata = fd;

        return kevent(epfd, &ev, 1, NULL, 0, NULL);
    }
    else if (EX_EPOLL_CTL_DEL == op)
    {
        NDRX_LOG(log_info, "%s: Delete operation on ndrx_epoll set %d, fd %d", fn, epfd, fd);

        EV_SET(&ev, real_fd, EVFILT_READ, EV_DELETE, 0, 0, NULL);

        return kevent(epfd, &ev, 1, NULL, 0, NULL);
    }
    else
    {
        NDRX_LOG(log_error, "Invalid operation %d", op);
        errno = EINVAL;
        EXFAIL_OUT(ret);
    }

out:

    NDRX_LOG(log_info, "%s return %d", fn, ret);

    return ret;

}

/**
 * Wrapper for epoll_create
 * @param size
 * @return 
 */
expublic inline int ndrx_epoll_create(int size)
{
    return kqueue();
}

/**
 * Close Epoll set.
 */
expublic inline int ndrx_epoll_close(int epfd)
{
    return close(epfd);
}

/**
 * Wrapper for epoll_wait
 * @param epfd
 * @param events
 * @param maxevents
 * @param timeout
 * @param buf not used
 * @param buf_len not used
 * @return 
 */
expublic int ndrx_epoll_wait(int epfd, struct ndrx_epoll_event *events,
        int maxevents, int timeout, char *buf, int *buf_len)
{
    int ret = EXSUCCEED;
    int err_saved;
    int numevents = 0;
    char *fn = "ndrx_epoll_wait";
    int i;
    int retpoll;
    struct kevent tevent;	 /* Event triggered */
    struct timespec tout;
    
    *buf_len = EXFAIL;

    if (timeout>0)
    {
    	NDRX_LOG(log_debug, "about to kevent wait %d, timeout %d", epfd, timeout);
        tout.tv_nsec = 0;
        tout.tv_sec = timeout/1000;
        retpoll = kevent(epfd, NULL, 0, &tevent, 1, &tout);
    }
    else
    {
    	NDRX_LOG(log_debug, "about to kevent wait %d, no timeout", epfd);
        retpoll = kevent(epfd, NULL, 0, &tevent, 1, NULL);
    }
    
    err_saved = errno;
    
    
    if (retpoll > 0)
    {
        /* 1 event currently supported... */
        NDRX_LOG(log_debug, "fd=%d, i=%d revents=%u", 
                tevent.ident, tevent.flags);

        numevents++;

        if (tevent.udata!=NULL)
        {
            /* the mqueue hanlder is encoded here (ptr to fd..)*/
            events[numevents-1].data.mqd = (mqd_t)tevent.udata;
            events[numevents-1].is_mqd = EXTRUE;
        }
        else
        {
           events[numevents-1].data.fd = tevent.ident;
           events[numevents-1].is_mqd = EXFALSE;
        }
        /* will provide back both flags & flag flags... */
        events[numevents-1].events = tevent.flags;
    }
    else if (retpoll==0)
    {
        goto out;
    }
    else
    {
        EXFAIL_OUT(ret);
    }
    
out:

    NDRX_LOG(log_info, "%s ret=%d numevents=%d", fn, ret, numevents);

    if (EXSUCCEED==ret)
    {
        return numevents;
    }
    else
    {
        errno = err_saved;
        return EXFAIL;
    }
}


/**
 * Return errno for ndrx_poll() operation
 * @return 
 */
expublic inline int ndrx_epoll_errno(void)
{
    return errno;
}

/**
 * Wrapper for strerror
 * @param err
 * @return 
 */
expublic inline char * ndrx_poll_strerror(int err)
{
    return strerror(err);
}

/**
 * not used by Kqueue epoll()
 * @param idx
 * @return 
 */
expublic int ndrx_epoll_shallopenq(int idx)
{
    return EXTRUE;
}
/**
 * Not used by Kqueue epoll()
 * @param qstr
 */
expublic void ndrx_epoll_mainq_set(char *qstr)
{
    return;
}

/**
 * Not used by Kqueue epoll
 * @param svcnm
 * @param idx
 * @param mq_exits
 * @return 
 */
expublic mqd_t ndrx_epoll_service_add(char *svcnm, int idx, mqd_t mq_exits)
{
    return mq_exits;
}

/**
 * Not used by linux epoll
 * @return 
 */
expublic int ndrx_epoll_shmdetach(void)
{
    return EXSUCCEED;
}

/**
 * Not used by Kqueue epoll
 * @param force
 * @return 
 */
expublic int ndrx_epoll_down(int force)
{
    return EXSUCCEED;
}

/* vim: set ts=4 sw=4 et smartindent: */
